module HAL where

import Conduit ( yieldMany, concatMapMC, takeC, (.|), ConduitT )
import Control.Monad.Fail (fail)
import Data.Aeson ( FromJSON )
import Data.LanguageCodes (ISO639_1(..))
import Data.Text qualified as T
import HAL.Client ( SortField(Asc), search, structure, searchCursor )
import HAL.Doc.Corpus ( Corpus )
import HAL.Doc.Struct ( Struct )
import HAL.Types (Response(..))
import HAL.Utils (langAbstractS, toText)
import Network.HTTP.Client (newManager, Request)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Protolude
import Servant.API ( ToHttpApiData )
import Servant.Client (BaseUrl(..), Scheme(..), ClientM, ClientError, runClientM, mkClientEnv, ClientEnv (makeClientRequest))
import System.IO.Unsafe (unsafePerformIO)


data HalCrawlerOptions
  = HalCrawlerOptions
  { -- | If 'True', enable the debug logs to stdout.
    _hco_debugLogs :: !Bool
  , _hco_batchSize :: !Int
  }

defaultHalOptions :: HalCrawlerOptions
defaultHalOptions = HalCrawlerOptions
  { _hco_debugLogs = False
  , _hco_batchSize = 300
  }

type Query = Text
type Start = Int
type Limit = Integer
type Count = Integer

queryWithLang :: Maybe ISO639_1 -> [Query]
queryWithLang Nothing = []
queryWithLang (Just lang) = ["language_s:" <> toText lang]


getMetadataWithCursorC :: Query
                       -- ^ The textual query
                       -> Maybe Limit
                       -- ^ An optional limit for the search, it influences the number of
                       -- rows returned.
                       -> Maybe ISO639_1
                       -- ^ An optional language for the search.
                       -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithCursorC = getMetadataWithCursorOptsC defaultHalOptions

-- | Fetch metadata using cursors
-- https://api.archives-ouvertes.fr/docs/search#cursors
getMetadataWithCursorOptsC :: HalCrawlerOptions
                           -- ^ The options for the crawler
                           -> Query
                           -- ^ The textual query
                           -> Maybe Limit
                           -- ^ An optional limit for the search, it influences the number of
                           -- rows returned.
                           -> Maybe ISO639_1
                           -- ^ An optional language for the search.
                           -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithCursorOptsC opts@HalCrawlerOptions { .. } q mb_limit lang = do
  -- Basically this works as follows:
  -- - fetch first page with cursor = "*"
  -- - get next cursor from the results
  -- - feed the cursor to get next page
  -- - when previous and current cursors are equal, there are no more results

  -- First, estimate the total number of documents
  eCount <- countResultsOpts' opts q lang
  pure $ get' <$> eCount
  where
    sort_ = Just $ Asc "docid"
    fq = queryWithLang lang

    get' :: Count
         -> (Maybe Count, ConduitT () Corpus IO ())
    get' numFound' =
      ( Just numResults
      , producer "*"
        -- | we need takeC again, because getPage could give too many results
        .| takeC (fromIntegral numResults)
      )
      where
        limit  = min numFound' $ fromMaybe numFound' mb_limit
        numResults = limit

    producer :: Text -> ConduitT () Corpus IO ()
    producer cursor = do
      let endpoint = searchCursor (Just q) (Just $ requestedFields lang) fq sort_ (Just $ fromIntegral _hco_batchSize) (Just cursor)
      liftIO $ debugLog opts $ "[getMetadataWithCursorLangC] producer: " <> show cursor
      eRes <- liftIO $ runHalAPIClient opts endpoint
      case eRes of
        Left err -> fail $ "error: " <> show err
        Right (Response { _docs, _nextCursorMark }) -> do
          yieldMany _docs
          case _nextCursorMark of
            Nothing -> fail "Expected next cursor mark, but got nothing"
            Just nextCursor -> do
              if cursor == nextCursor then
                pure ()
              else do
                producer nextCursor


countResults' :: Query -> Maybe ISO639_1 -> IO (Either ClientError Count)
countResults' q lang = do
  countResultsOpts' defaultHalOptions q lang

countResultsOpts' :: HalCrawlerOptions -> Query -> Maybe ISO639_1 -> IO (Either ClientError Count)
countResultsOpts' opts q lang = do
  -- Set rows=0 to query number of results
  -- https://api.archives-ouvertes.fr/docs/search#rows
  -- First, estimate the total number of documents
  eRes <- runHalAPIClient opts $ search (Just q) (Just $ requestedFields Nothing) fq Nothing (Just 0) (Just 0) :: IO (Either ClientError (Response Corpus))
  pure (fromIntegral . _numFound <$> eRes)
  where
    fq = queryWithLang lang



--- SOME OTHER FETCH FUNCTIONS


getMetadataWith :: [Query]
                 -- ^ The textual query
                -> Maybe Start
                 -- ^ An optional offset
                -> Maybe Limit
                -- ^ An optional limit for the search, it influences the number of
                -- rows returned.
                -> Maybe ISO639_1
                -- ^ An optional language for the search.
                -> IO (Either ClientError (Response Corpus))
getMetadataWith qs start_ limit lang = do
  runHalAPIClient defaultHalOptions $ search (Just q) (Just $ requestedFields lang) [] Nothing start_ (fromIntegral  <$> limit)
  where
    q = joinQueries $ qs <> queryWithLang lang

-- | Fetch results, returning a Conduit stream.
--   NOTE: Prefer fetching with `getMetadataWithCursorC` instead of this function.
getMetadataWithC :: [Query]
                 -- ^ The textual query
                 -> Maybe Start
                 -- ^ An optional offset
                 -> Maybe Limit
                 -- ^ An optional limit for the search, it influences the number of
                 -- rows returned.
                 -> Maybe ISO639_1
                 -- ^ An optional language for the search.
                 -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithC qs start_ limit lang = getMetadataWithOptsC defaultHalOptions (qs <> queryWithLang lang) start_ limit lang

getMetadataWithOptsC :: HalCrawlerOptions
                     -- ^ The options for the crawler
                     -> [Query]
                     -- ^ The textual query
                     -> Maybe Start
                     -- ^ An optional offset
                     -> Maybe Limit
                     -- ^ An optional limit for the search, it influences the number of
                     -- rows returned.
                     -> Maybe ISO639_1
                     -- ^ An optional language for the search.
                     -> IO (Either ClientError (Maybe Count, ConduitT () Corpus IO ()))
getMetadataWithOptsC opts@HalCrawlerOptions { .. } qs mb_offset mb_limit lang = do
  -- First, estimate the total number of documents
  eCount <- countResults qs
  pure $ get' <$> eCount
  where
    get' :: Count
         -> (Maybe Count, ConduitT () Corpus IO ())
    get' numFound' =
      ( Just numResults
      , yieldMany [0..]
        .| takeC (fromIntegral numPages)
        .| concatMapMC (getPage offset)
        -- | we need takeC again, because getPage could give too many results
        .| takeC (fromIntegral numResults)
      )
      where
        offset = fromMaybe 0 mb_offset
        limit  = min numFound' $ fromMaybe numFound' mb_limit
        numResults = limit - fromIntegral offset
        numPages = numResults `div` fromIntegral _hco_batchSize + 1

    getPage :: Start -> Int -> IO [Corpus]
    getPage start' pageNum = do
      let offset = start' + pageNum * _hco_batchSize
      debugLog opts $ "[getMetadataWithLangC] getPage: " <> show offset
      eRes <- runHalAPIClient opts $ search (Just q) (Just $ requestedFields lang) [] Nothing (Just offset) (Just $ fromIntegral _hco_batchSize)
      pure $ case eRes of
        Left _ -> []
        Right Response { _docs } -> _docs
      where
        q = joinQueries qs


debugLog :: HalCrawlerOptions -> Text -> IO ()
debugLog HalCrawlerOptions{..} msg = when _hco_debugLogs $ putStrLn msg

countResults :: [Query] -> IO (Either ClientError Count)
countResults qs = do
  -- Set rows=0 to query number of results
  -- https://api.archives-ouvertes.fr/docs/search#rows
  -- First, estimate the total number of documents
  eRes <- runHalAPIClient defaultHalOptions $ search (Just q) (Just $ requestedFields Nothing) [] Nothing (Just 0) (Just 0) :: IO (Either ClientError (Response Corpus))
  pure (fromIntegral . _numFound <$> eRes)
  where
    q = joinQueries qs





requestedFields :: Maybe ISO639_1 -> Text
requestedFields (Just EN) = T.intercalate "," baseFields
requestedFields (Just lang) = T.intercalate "," $ baseFields <> [langAbstractS lang]
 -- "docid,title_s,en_abstract_s," <> langAbstractS lang <> ",submittedDate_s,source_s,authFullName_s,authOrganism_s"
requestedFields _ = requestedFields (Just EN)

baseFields :: [Text]
baseFields = [ "docid"
             , "title_s"
             , "en_abstract_s"
             , "submittedDate_s"
             , "source_s"
             , "authFullName_s"
             , "authOrganism_s"
             , "structIdName_fs" ]

structFields :: Text
structFields = "docid,label_s,parentDocid_i"

requestLog :: HalCrawlerOptions -> Request -> Request
requestLog opts rq = unsafePerformIO $ do
  debugLog opts $ "[HAL.makeClientRequestLog] " <> show rq
  pure rq
{-# NOINLINE requestLog #-}

runHalAPIClient :: (FromJSON doc, ToHttpApiData doc)
                => HalCrawlerOptions
                -> ClientM (Response doc)
                -> IO (Either ClientError (Response doc))
runHalAPIClient opts cmd = do
  manager' <- newManager tlsManagerSettings
  runClientM cmd' (mkClientEnv manager' $ BaseUrl Https "api.archives-ouvertes.fr" 443 "")
  where
    cmd' = local (\r -> r {
      makeClientRequest = \bUrl servantRq -> requestLog opts (makeClientRequest r bUrl servantRq)
      }) cmd

runStructureRequest :: Maybe Text -> IO (Either ClientError (Response Struct))
runStructureRequest rq =
  runHalAPIClient defaultHalOptions $ structure (Just structFields) rq (Just 10000)

runSearchRequest :: [Text] -> IO (Either ClientError (Response Corpus))
runSearchRequest rq =
  runHalAPIClient defaultHalOptions $ search (Just q) (Just $ requestedFields Nothing) [] Nothing Nothing Nothing
  where
    q = joinQueries rq

generateRequestByStructID :: Text -> [Text] -> Text
generateRequestByStructID rq struct_ids =
  "text:" <>
  rq
  <> " AND "
  <> "structId_i:("
  <> flattenPipe struct_ids
  <> ")"

flattenPipe :: [Text] -> Text
flattenPipe [] = ""
flattenPipe [x] = x
flattenPipe (x:xs) = x <> " || " <> flattenPipe xs


joinQueries :: [Text] -> Text
joinQueries = T.intercalate " AND "
