{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DeriveGeneric #-}

module HAL.Doc.Corpus where

import Control.Lens qualified as L
import Data.Aeson ( (.:), (.:?), withObject, FromJSON(parseJSON), Object )
import Data.Default (Default(..))
import Data.Map.Strict qualified as Map
import Data.String (IsString(fromString))
import HAL.Utils (allLangs, langAbstractS)
import Data.LanguageCodes (ISO639_1(..))
import Protolude
import Servant.API (ToHttpApiData(..))
import qualified Data.Text as T
import Data.Aeson.Types (Parser)
import Prelude qualified as P

data Corpus = Corpus
  { _corpus_docid                :: Text
  , _corpus_title                :: [Text]
  , _corpus_abstract             :: [Text]
  , _corpus_abstract_lang_map    :: Map ISO639_1 [Text]
  , _corpus_date                 :: Maybe Text
  , _corpus_source               :: Maybe Text
  , _corpus_authors_names        :: [Text]
  , _corpus_authors_affiliations :: [Text]
  , _corpus_struct_id            :: [Int]
  , _corpus_original             :: Object
  } deriving (Show, Generic)
L.makeLenses ''Corpus

instance Default Corpus where
  def = Corpus "default Id" def def def def def def def def mempty

instance FromJSON Corpus where
  parseJSON = withObject "Corpus" $ \o -> do
    _corpus_docid                <- o .: "docid"
    _corpus_title                <- o .: "title_s" <|> return []
    _corpus_abstract             <- o .: "en_abstract_s" <|> return []
    _corpus_date                 <- o .:? "submittedDate_s"
    _corpus_source               <- o .:? "source_s"
    _corpus_authors_names        <- o .: "authFullName_s" <|> return []

    idsNames <- o .:? "structIdName_fs" :: Parser (Maybe [Text]) --unsparsed (contains a _FacetSep_)

    let structIdname = getStructIdsNames idsNames

    let _corpus_struct_id = map fst structIdname
    let _corpus_authors_affiliations = map snd structIdname

    abstracts <-
      mapM (\lang -> do
               ma <- o .:? fromString (T.unpack $ langAbstractS lang)
               pure $ (\a -> (lang, a)) <$> ma) allLangs
    let _corpus_abstract_lang_map = Map.fromList $ catMaybes abstracts

    let _corpus_original = o

    pure $ Corpus { .. }

-- | this function parses the field structIdName_fs that looks like : 
-- > StructId_FacetSep_StructName
-- 
-- returns [(StructId, StructName)]
getStructIdsNames :: Maybe [T.Text] -> [(Int, T.Text)]
getStructIdsNames (Just idsNames) = map (\tab -> (P.read (T.unpack (P.head tab)) :: Int, P.last tab)) $ splitInstitutes idsNames
  where
    splitInstitutes :: [T.Text] -> [[T.Text]]
    splitInstitutes = P.map (T.splitOn (T.pack "_FacetSep_"))
getStructIdsNames Nothing = []

instance ToHttpApiData Corpus where
  toUrlPiece _ = "docid,title_s,en_abstract_s,fr_abstract_s,submittedDate_s,source_s,authFullName_s,structId_i,structIdName_fs"
